package com.arm.android.example.dalvikmcpkg;

import android.app.Activity;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.os.AsyncTask;
import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.os.SystemClock;
import android.view.View;
import android.widget.Button;
import android.widget.ImageView;
import android.widget.TextView;

import java.util.concurrent.Semaphore;

public class DalvikmcActivity extends Activity {
	// Widgets and resources
	private   static Button[]    BN;
	private   static TextView[]  TV;
	private   static ImageView[] IV;
	private   static int[]       imgResID;
	
	// State information
	protected int numCores;
	private   int imgSequence;
	
	// Synchronization and message passing
	private   Semaphore decodeLock;
	private   Semaphore brotSignal;
	private   Handler   UIThandler;
	final     int       MSG_SETTEXTVIEW  = 1;
	final     int       MSG_SETIMAGEVIEW = 2;
	
	protected class mandelbroter implements Runnable {
		private int brotiv;
		private int brottv;

		mandelbroter(int iv, int tv)
		{
			brotiv = iv; // Target ImageView
			brottv = tv; // Target TextView (for logging)
		}

		/*
		 * Send a message to UI thread to update our associated TextView
		 */
		private void updateTextView(String str)
		{
			Message mymsg = new Message();
			mymsg.what = MSG_SETTEXTVIEW;
			mymsg.arg1 = brottv;
			mymsg.obj = (Object) str;
			
			UIThandler.sendMessage(mymsg);
		}
		
		/*
		 * Send a message to UI thread to update our associated ImageView
		 */
		private void updateImageView(Bitmap bmp)
		{
			Message mymsg = new Message();
			mymsg.what = MSG_SETIMAGEVIEW;
			mymsg.arg1 = brotiv;
			mymsg.obj = (Bitmap) bmp;
			
			UIThandler.sendMessage(mymsg);
		}

		/*
		 * Adapted from well documented example at http://warp.povusers.org/Mandelbrot/
		 */
		private void genMandelbrot(int[] buf, int width, int height, int colourscheme)
		{
			double MinRe = -2.0;
			double MaxRe = 1.0;
			double MinIm = -1.2;
			double MaxIm = MinIm + (MaxRe - MinRe) * height / width;
			double Re_factor = (MaxRe - MinRe) / (width - 1);
			double Im_factor = (MaxIm - MinIm) / (height - 1);
			int x, y, isInside, n;

			for(y = 0 ; y < height ; y++) {
			    double c_im = MaxIm - (y * Im_factor);
			    
			    for(x = 0 ; x < width ; x++) {
			        double c_re = MinRe + x*Re_factor;
			        double Z_re = c_re;
			        double Z_im = c_im;

			        isInside = 1;

			        // I think 32 iterations give the prettiest picture :)
			        for(n = 0; n < 32; n++) {
			            double Z_re2 = Z_re*Z_re;
			            double Z_im2 = Z_im*Z_im;

			            if(Z_re2 + Z_im2 > 4) {
			                isInside = 0;
			                break;
			            }

			            Z_im = 2*Z_re*Z_im + c_im;
			            Z_re = Z_re2 - Z_im2 + c_re;
			        }

			        // Scale 0-32 => 0-255
		        	n = n * 8;
		        	if (n == 256)
		        		n = 255;
		        	
		        	// Set current pixel value
			        if(isInside == 1) {
			        	// Make members of the set black
			        	buf[(y * width) + x] = 0xff000000;
			        } else {
			        	// Make non-members pretty colours
			        	int rshift, gshift, bshift;
			        	
			        	switch (colourscheme) {
			        	case 0: // Red
			        		rshift = 0;
			        		gshift = 1;
			        		bshift = 2;
			        		break;
			        	case 1: // Green
			        		rshift = 2;
			        		gshift = 0;
			        		bshift = 1;
			        		break;
			        	case 2: // Blue
			        		rshift = 1;
			        		gshift = 2;
			        		bshift = 0;
			        		break;
			        	default:
			        		rshift = gshift = bshift = 0;
			        	}
			        	
			        	// Assign weighted colour values
			        	int rval = (n >> rshift) < 1 ? 1 : (n >> rshift) & 0xff;
			        	int gval = (n >> gshift) < 1 ? 1 : (n >> gshift) & 0xff;
			        	int bval = (n >> bshift) < 1 ? 1 : (n >> bshift) & 0xff;

			        	// Alpha | Red | Green | Blue => ARGB_8888 for Android Bitmap
			        	int pixel = (0xff << 24) | (rval << 16) | (gval << 8) | (bval << 0);
			        	buf[(y * width) + x] = pixel;
			        }
			    }
			}
		}
		
		private void drawMandelbrot(int colourscheme)
		{
			// Find out how much screen real estate we have to work with
			int height = IV[brotiv].getHeight();
			int width  = IV[brotiv].getWidth();

			// Take a timestamp for metrics
			long starttime = SystemClock.elapsedRealtime();

			// Reserve buffer and fill it
			int[] buf = new int[height*width];
			genMandelbrot(buf, width, height, colourscheme);
			
			// Convert resulting int array to a Bitmap object
			Bitmap bmp = Bitmap.createBitmap(buf, width, height, Bitmap.Config.ARGB_8888);

			/* Get end time reading for metric */
            long endtime = SystemClock.elapsedRealtime();

            // Send decoded bitmap to UI thread
            updateImageView(bmp);
			
            // Print processing time
            updateTextView("fractal time=" + (endtime - starttime));
		}
		
		// Mandelbroter main thread function
		public void run()
		{
			int colourscheme=0;

			while(true) {
				// Wait for signal to generate image
				brotSignal.acquireUninterruptibly();

				drawMandelbrot(colourscheme);

				// Cycle through available colour schemes
				colourscheme = (colourscheme + 1) % 3;
			}
		}
	}
	
	private class ivUpdater extends AsyncTask<Object, Integer, Bitmap> {
		private long starttime, endtime;
		private int decodeiv, decodetv;
		
		ivUpdater(int iv, int tv)
		{
			decodeiv = iv; // Target ImageView
			decodetv = tv; // Target TextView (for logging)
		}
		
		/* Take a resource ID, return a Bitmap created from it, including the JPG decompression */
		protected Bitmap getBitmapFromRes(int ID)
		{
			BitmapFactory.Options opts = new BitmapFactory.Options();
			return BitmapFactory.decodeResource(getResources(), ID, opts);
		}
		
		/* Executed on non-UI thread */
		protected Bitmap doInBackground(Object... o)
		{
			/* Ensure only one instance of handler modifies the common objects at once */
			decodeLock.acquireUninterruptibly();
			
			/* Take a timestamp for metrics */
			starttime = SystemClock.elapsedRealtime();

			/* Decode next image in sequence */
			Bitmap bmp = getBitmapFromRes(imgResID[imgSequence]);
            
			/* Get end time reading for metric */
            endtime = SystemClock.elapsedRealtime();
            
            /* Update sequence counter */
            imgSequence = (imgSequence + 1) % imgResID.length;
			
            /* Pass the array of bitmaps to onPostExecute() */
            return bmp;
		}

		/* Executed on UI thread once doInBackground returns, receiving its output */
		protected void onPostExecute(Bitmap bmp) {
			/* Update each ImageView */
			for (int i=0 ; i < IV.length ; i++)
				IV[decodeiv].setImageBitmap(bmp);

			/* And since we are on UI thread, we can update text directly */
			TV[decodetv].setText("image time=" + (endtime - starttime));
			
			/* Permit new button handlers to be invoked */
        	decodeLock.release();
		}
	}
	
    /** Called when the activity is first created. */
    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.main);
        
        /* 
         * Look up IDs of my widgets and resources.
         */
        BN = new Button[] {
        		(Button) findViewById(R.id.decodebutton),
        		(Button) findViewById(R.id.brotbutton),
        		(Button) findViewById(R.id.bothbutton),
        };

        TV = new TextView[] {
        		(TextView) findViewById(R.id.statusbox1),
        		(TextView) findViewById(R.id.statusbox2)
        };

        IV = new ImageView[] {
        				(ImageView) findViewById(R.id.image1),
        				(ImageView) findViewById(R.id.image2)
        			};
        imgResID = new int[] {
				R.drawable.dsc_0358,
				R.drawable.dsc_0360
			};
        
        // Find out how many cores we have to work with - might come in handy
    	numCores = Runtime.getRuntime().availableProcessors();

    	// Write something cute in the TextViews
    	TV[0].setText("Number of cores: " + numCores);
    	TV[1].setText("Number of cores: " + numCores);
		
		// Semaphore to prevent simultaneous decodes to the same ImageView
    	decodeLock = new Semaphore(1);
    	
    	// Semaphore to manually trigger mandelbrot generation
    	brotSignal = new Semaphore(0);

    	// Define "Decode" button handler inline
        BN[0].setOnClickListener(new View.OnClickListener() {
            public void onClick(View v) {
            	/* 
            	 * Do the decoding explicitly using the thread pool executor.
            	 */
           		new ivUpdater(0, 0).executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR);
            }
        });
        
    	// Define "Brot" button handler inline
        BN[1].setOnClickListener(new View.OnClickListener() {
            public void onClick(View v) {
            	/* 
            	 * Trigger a single mandelbrot generation
            	 */
           		brotSignal.release();
            }
        });

    	// Define "Both" button handler inline
        BN[2].setOnClickListener(new View.OnClickListener() {
            public void onClick(View v) {
            	/* 
            	 * Trigger simultaneous decode and mandelbrot
            	 */
           		new ivUpdater(0, 0).executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR);
           		brotSignal.release();
            }
        });

        // Set up handler for updating textview
        UIThandler = new Handler(){
        	public void handleMessage(Message msg)
        	{        		
        		switch (msg.what) {
        		case MSG_SETTEXTVIEW:
        			// Write String in "obj" into specified TextView 
        			TV[msg.arg1].setText((String) msg.obj);
        			break;
        		case MSG_SETIMAGEVIEW:
        			//  Set Bitmap in "obj" as content of specified ImageView
        			IV[msg.arg1].setImageBitmap((Bitmap)msg.obj);
        			break;
        		}
        	}
        };
    }
    
    protected void onStart()
    {
        super.onStart();

        /* 
         * Launch mandelbroter background thread rendering to IV[1]
         * and writing status messages to TV[1]
         */
        Thread t = new Thread(new mandelbroter(1, 1));
        t.start();
    }
}